;; Datalog syntax parser
;; version1.2.0 
;; Fri 02 Apr 2004 11:04:07 AM MSD
;; it is written to Bigloo specially

; This is Datalog LALR(1) grammar
; Exported:
;
; dtlog-syntax
; dtlog-lex
; dtlog-parse

; replace-avariable 
; 
; my-bind
; parse-identifier
; parse-number
; parse-operation
; parse-assign
; parse-aggregate
;
; get-idb-names
;
;==============================================================================
; Datalog parser
; 
(define rule-tag 'r)
(define query-tag 'q)
(define instr-tag 'i)
(define db-tag 'db)
(define anonymous-tag '_)
(define sum-tag 'sum)
(define max-tag 'max)

; replace-avariable :: rule -> new rule
; replace-avariable replaces all anonymous variables to usual variable
(define (replace-avariable rule)
  (define used-var #f)
  (map (lambda (atom) 
    (if (atom? atom)
      (map (lambda (term-const-pred)
             (if (var? term-const-pred)
   	         (if (equal? term-const-pred anonymous-tag)
                     (let ((new-var (newvar (if used-var
	                                      (list used-var)
	 	    	   	  	      (filter var? (flat-rule-terms rule))))))
                        (set! used-var new-var)
         	        new-var)   
	             term-const-pred)
	         term-const-pred)) atom) atom)) rule))
	       
;(cout (replace-avariable '((p1 _ X Z) (p2 X3 Y) (p3 X1 Z _))) nl)
; -> ((p1 X4 X Z) (p2 X3 Y) (p3 X1 Z X5))  
;(cout (replace-avariable '((p1 "12" _X1 _ _X2))) nl)
; -> ((p1 12 _X1 X1 _X2))


; my pretty-print
(define (local-pp var . vars)
  (cout var) 
  (map (lambda (x) (cout x " ")) vars)
  (cout nl) var)

; my-bind :: names -> vals -> name -> val
; my-bind finds index name in list names and return vals[this index]
; vals is list too
(define (my-bind names vals name)
  (list-ref vals (list-index (lambda (x) (equal? x name)) names)))

; parse-identifier :: identifier -> internal-data
; returns the internal presentation of identifier
(define (parse-identifier identifier)
  (list (list identifier)
        (lambda (idents vals) 
;	    (cout idents nl vals nl identifier nl)
    (let ((v (my-bind idents vals identifier)))
      (value->number v)))))
      
; parse-number :: number -> internal-data
; returns the internal presentation of number
(define (parse-number number)
  (list '() (lambda (idents vals) number)))
 
; parse-operation :: op -> expr1 -> expr2 -> internal-data
; returns the internal presentation of operation
(define (parse-operation op expr1 expr2)
 (list (append (car expr1) (car expr2))
       (lambda (idents vals) 
         (op ((cadr expr1) idents vals)
      ((cadr expr2) idents vals)))))

; parse-assign :: identifier -> expr -> internal-data
; returns the internal presentation of assign predicate
(define (parse-assign identifier expr)
  (append expr (list identifier)))

; parse-aggregate :: id1 -> id2 -> predicate-list -> aggr-function -> internal-data
; returns the internal presentaion of aggregate predicate
(define (parse-aggregate id1 id2 predicate-list aggr-function)
  (list '() eq? id1 id2 predicate-list aggr-function)) 

; dtlog-syntax is a variable contained lalr(1) grammar of Datalog
(define dtlog-syntax
 (lalr-grammar 
  ((left: op-mul op-div) (left: op-add op-sub)
   lb rb lbb rbb em! dot q? comma conseq identifier constant db str _ number
	eq neq less less-eq gr gr-eq assign sum max)
  (program
    ((statement-list)			statement-list))
  (statement-list
    ((statement)			(list statement))
    ((statement-list statement)		(append statement-list (list statement))))
  (statement
    ((rule)				(list rule-tag (replace-avariable rule)))
    ((query)				(list query-tag query))
    ((processing-instruction)    	(list instr-tag processing-instruction)))
  (rule
    ((predicate conseq predicate-list dot) 	
					(append (list predicate) predicate-list)))
  (query
    ((q? predicate-list dot) 		(append '(()) predicate-list)))        
  (processing-instruction
    ((em! instruction)			instruction))
  (instruction
    ((db str)				(list db-tag str)))
  (predicate-list
    ((predicate) 			(list predicate))
    ((predicate-list comma predicate)	(append predicate-list (list predicate))))
 (predicate
    ((eval-expr)			eval-expr)
    ((constant lb term-list rb) 	(append (list (string->symbol constant)) term-list)))
  (term-list
    ((term)				term)
    ((term-list comma term)		(append term-list term)))
  (term
    ((str)				(list str))
    ((constant)				(list constant))
    ((variable)				(list variable)))
  (variable
    ((identifier)			identifier)
    ((_)				_))
  (eval-expr
    ((assign-expr)			assign-expr)
    ((logic-pred)			logic-pred))
  (logic-pred
    ((ar-expr@expr1 rel-sign ar-expr@expr2)
					(parse-operation rel-sign expr1 expr2)))
  (ar-expr
    ((ar-expr@expr1 ar-op ar-expr@expr2)
					(parse-operation ar-op expr1 expr2))
    ((lb ar-expr rb)    		ar-expr)
    ((identifier)			(parse-identifier identifier))
    ((number)				(parse-number number)))
  (rel-sign
    ((eq) eq) ((neq) neq) ((less) less) ((less-eq) less-eq) ((gr) gr) ((gr-eq) gr-eq))
  (ar-op
    ((op-mul) op-mul) ((op-div) op-div) ((op-add) op-add) ((op-sub) op-sub))
  (assign-expr
    ((identifier assign ar-expr@expr)	(parse-assign identifier expr))
    ((identifier@id1 assign aggr-function lb lbb predicate-list rbb comma identifier@id2 rb)
                                        (parse-aggregate id1 id2 predicate-list aggr-function)))
  (aggr-function
    ((sum) sum) ((max) max))					
))

(define line 1)
(define pos  0)

; dtlog-lex is a variable contained regular grammar for Datalog syntax
(define dtlog-lex
 (regular-grammar ()
  ((+ (or #\Newline #a013)) 	 (set! line (+ line (the-length))) (ignore))
  ((+ (or #\tab #\space)) 	 (ignore))
  ((eol (: #\; (* all)))                (ignore))
  ((eof (: #\; (* all)))                (ignore))
  ((uncase "db")			'db)
  ((: (or upper #\_) (* (or #\_ alnum))) (cons 'identifier (string->symbol (the-string))))
  ((: lower (* (or #\_ alnum)))			(cons 'constant (the-string)))
  ((+ digit)				(cons 'number (string->number (the-string))))
  ((: #\" (* (out #\Newline #\")) #\")
	            	                (cons 'str (substring (the-string) 1 (- (string-length (the-string)) 1))))					
  (#\_  (cons '_ anonymous-tag))
  ("!=" (cons 'neq (lambda (x y) (not (eq? x y)))))
  (#\!  'em!)
  (#\?  'q?)
  (#\(	'lb)
  (#\)  'rb)
  (#\[  'lbb)
  (#\]  'rbb)
  (#\.	'dot)
  (#\,  'comma)
  (":-" 'conseq)
  ("+"  (cons 'op-add +))
  ("-"  (cons 'op-sub -))
  ("*"  (cons 'op-mul *))  
  ("/"  (cons 'op-div /))  
  ("==" (cons 'eq eq?))
  ("<"  (cons 'less <))
  ("<=" (cons 'less-eq <=))
  (">"  (cons 'gr >))
  (">=" (cons 'gr-eq >=))
  ("="  'assign)
  ("sum!" (cons 'sum sum-tag))
  ("max!" (cons 'max max-tag))
  (else	(let ((char (the-failure)))
;          (cout nl char nl)
;	    char))))
          (if (eof-object? char)
	      char
	      (error "Datalog lex parser" "Illegal char " (char->integer char)))))))

; dtlog-parse :: filename -> internal dtlog-program
; dtlog-parse parses filename and returns internal dtlog-program
(define (dtlog-parse filename)
  (set! line 1)
  (set! pos 0)
  (define (parse-handler escape proc mes obj)
   (cout "Datalog syntax error! line#: " line nl proc " : " mes " = " obj nl)
   (escape #f))
  (let* ((port 
	(let ((port (open-input-file filename)))
	  (if port port (begin (cout "Could not open file: " filename nl) (exit -1)))))
         (program (try (read/lalrp dtlog-syntax dtlog-lex port) parse-handler)))
;    (cout program nl)
    (close-input-port port)
    program))
;(dtlog-parse "example")

; get-idb-names -> rules -> idb-names
(define (get-idb-names rules)
  (delete-duplicates (map caar rules)))
