(module dtlog-tools
    (import sqlite)
    (import os)
    (import times)
    (include "../common/myenv.scm")
    (include "../common/bigloo/srfi-1.scm")
    (include "../common/bigloo/common-bigloo.scm")
    (include "./bench-tools.scm")
    (include "./rel-algebra.scm")
    (include "./unify.scm")
    (include "./qsqr.scm")
    (include "./calculating.scm")
    (include "./dtlog-to-ra.scm")
    (include "./dtlog-parse.scm")
    (include "./dtlog-system.scm")
    (main main)
)

(define (usage)
 (cout "usage: dtlog-tools filename max-count-facts max-int-value" nl)
)

(define (main argv)
  (define export #f)
  (define create #f)
  (define create_anc #f)
  (define create_b2 #f)
  (define db_sqlite #f)
  (define txt_file #f)
  (define zebra_opt #f)
  (define max_count_opt #f)
  (define max_int_value_opt #f)

  (args-parse (cdr argv) 
    (section "help")
     ((("-h" "--help")
       (help "Print this help message"))
       (args-parse-usage #f))
    (section "options")
     (("-export" ?db-sqlite ?txt-file 
      (help "Export database from sqlite file to txt-file"))
      (set! export #t) (set! db_sqlite db-sqlite)
      (set! txt_file txt-file))
     (("-create" ?db-sqlite ?info-file
      (help "Create database in sqlite file. In <info-file> there are info about database"))
      (set! create #t) (set! db_sqlite db-sqlite)
      (set! txt_file info-file))
     (("-c" ?db-sqlite ?max-count ?max-int-value
      (help "Create database in sqlite file. Max-count and max-int-value are parameters"))
	 (set! db_sqlite db-sqlite)
	 (set! max_count_opt max-count) (set! max_int_value_opt max-int-value))
     (("-anc" (help "Create ancestor database"))
	 (set! create_anc #t)) 
     (("-b2" (help "Create benchmark2 database"))
	 (set! create_b2 #t))  
     (("-czebra" ?db-sqlite 
      (help "Create database to zebra puzzle in sqlite file. "))
      (set! zebra_opt #t) 
      (set! db_sqlite db-sqlite))
    (else (begin (cout "Illegal argument `" else "'" nl)
                 (cout "Usage:" nl)
                 (cout "dtlog-tools [options]" nl))
       (args-parse-usage #f)))
   (cond (export (db-export db_sqlite txt_file))
	 (create (tools-create db_sqlite txt_file))
	 (create_anc (create-anc db_sqlite max_count_opt max_int_value_opt))
	 (create_b2 (create-b2 db_sqlite max_count_opt max_int_value_opt))
	 (zebra_opt (create-zebra-edb db_sqlite))
	 
	)
)

; cout-port :: w -> args -> unspec
; displays to port w all args
(define (cout-port w . args)
  (for-each (lambda (arg) (display arg w)) args))

; db-export :: db_sqlite -> txt_fileopt -> unspecified
; export facts from db_sqlite to txt_file
(define (db-export db_sqlite txt_file) 
 (receive (db-names db) 
   (read-db db_sqlite) 
   (let ((w (open-output-file txt_file)))
     (for-each (lambda (pred-name pred-val)
    		 (for-each (lambda (fact) 
		   (cout-port w pred-name 
			     (drop-right (fold (lambda (x res) (append res (list x ","))) '() 
			       (vector->list fact)) 1)
		      ".")) pred-val))
	        db-names db)
     (close-output-port w))))

(define (tools-create-info filename db-info)
  (create-db db-info filename)
; (create-db `(("par" 2 ,max-count-facts ,max-int-value)) filename)
  (let ((full-db (read-db filename)))
;    (cout "full-db = " full-db nl)
    (cout "Facts count in database " filename " is " (facts-count (map second full-db)) nl)
 ))

(define (tools-create filename info_file)
  (let* ((w (open-input-file info_file))
	 (db-info (read w)))
 (tools-create-info filename db-info)
 (close-input-port w)
 )
)

(define (create-b2 filename max-count-facts max-int-value)
  (tools-create-info filename 
		`(("c1" 2 ,max-count-facts ,max-int-value) 
                ("c2" 2 ,max-count-facts ,max-int-value)
		("c3" 2 ,max-count-facts ,max-int-value) 
		("c4" 2 ,max-count-facts ,max-int-value))))

(define (create-anc filename max-count-facts max-int-value)
  (tools-create-info filename `(("par" 2 ,max-count-facts ,max-int-value))))
;  (tools-create-info filename `(("par" 2 10 100))))

; it assumes the structure of db is already created
(define (create-zebra-edb filename)
  (cout "create zebra edb" nl)
  (let* ((conn (sqlite-connect filename))
         (sess (acquire conn))
         (ids '("1" "2" "3" "4" "5"))
         (tbls
  '(("natio" ("Norwegian" "Englishman" "Spaniard" "Ukrainian" "Japanese"))
    ("smoke" ("Kools" "Chesterfields" "Parliament" "LuckyStrike" "Winston"))
    ("pet" ("Dog" "Fox" "Snails" "Horse" "Zebra"))
    ("drink" ("Coffee" "Tea" "Milk" "Water" "OJ"))
    ("color" ("Red" "Blue" "Ivory" "Green" "Yellow")))))
  (for-each 
    (lambda (x)
      (for-each 
        (lambda (id)
          (for-each (lambda (val) (insert-values (car x) (list id val) conn sess))
	      (cadr x))) ids)) tbls)
  (for-each (lambda (id) (insert-values "member" (list id) conn sess)) ids)
  (cout "ok")
  )
)
             
