(module
  sqlite
  (option (loadq "common.sch"))
  ;(include "common.sch")
  (import (os "os.scm"))
  (extern
    (include "sqlite.h")
    (export
      sqlite-execute-callback
      "sqlite_execute_callback"))
  (export
    (class sqlite-connection
           impl::sqlite)
    (generic _dismiss! ::sqlite-connection)

    (class sqlite-session
	   connection::sqlite-connection
           (statement (default #f))
           (query (default #f))
           (result (default #f))
           (columns (default #f)))
	 (sqlite-connect::sqlite-connection filename::string)
	 (generic acquire::sqlite-session ::sqlite-connection)
	 (generic execute::bool ::sqlite-session)
	 (generic prepare::bool ::sqlite-session sql::bstring)
	 (generic bind! ::sqlite-session bindings::pair-nil)
	 (generic fetch!::pair-nil ::sqlite-session)
	 (generic describe::pair-nil ::sqlite-session)
	 (generic fetch-all!::pair-nil ::sqlite-session)
    (sqlite-execute-callback::bool
      this::void*
      argc::int
      argv::string*
      column-names::string*))
  (type (subtype sqlite "sqlite*" (cobj))
        (coerce cobj sqlite () (cobj->sqlite))
        (coerce sqlite cobj () (sqlite->cobj))
        (coerce
          sqlite
          bool
          ()
          ((lambda (x) (pragma::bool "$1 != NULL" x))))
        (subtype bsqlite "obj_t" (obj))
        (coerce obj bsqlite () ())
        (coerce bsqlite obj () ())
        (coerce
          bsqlite
          sqlite
          (sqlite?)
          (bsqlite->sqlite))
        (coerce
          sqlite
          obj
          ()
          ((lambda
             (result)
             (pragma::bsqlite
               "cobj_to_foreign($1, $2)"
               'sqlite
               result)))))
  (foreign
    (macro sqlite cobj->sqlite (cobj) "(sqlite*)")
    (macro cobj sqlite->cobj (sqlite) "(long)")
    (macro sqlite
           bsqlite->sqlite
           (foreign)
           "(sqlite*)FOREIGN_TO_COBJ"))
  (export (sqlite?::bool o::obj)))

;(register-eval-srfi! 'sqlite)

(define
  (sqlite?::bool o::obj)
  (and (foreign? o) (eq? (foreign-id o) 'sqlite)))

(define
  (sqlite-connect::sqlite-connection
    filename::string
;    #!optional
;    (mode 438)
)
  (let* ;((mode::int mode)
         ((mode::int 438)
         (errmsg::string (pragma::string "NULL"))
         (conn::sqlite
           (pragma::sqlite
             "sqlite_open($1, $2, &$3)"
             filename
             mode
             errmsg)))
        (if (pragma::bool "$1 == NULL" conn)
            (error "sqlite-connect" errmsg filename)
            (instantiate::sqlite-connection (impl conn)))))

;(rdbms-register! "sqlite" sqlite-connect)

(define-generic
  (acquire::sqlite-session self::sqlite-connection)
  (instantiate::sqlite-session (connection self)))

(define-generic
  (prepare::bool self::sqlite-session sql::bstring)
  (with-access::sqlite-session
    self
    (statement query result)
    (set! statement sql)
    (set! query #f)
    (set! result #f)
    #t))

(define
  (sqlite-execute-callback::bool
    this::void*
    argc::int
    argv::string*
    column-names::string*)
  (let ((this::sqlite-session
          (pragma::sqlite-session "$1" this)))
       (with-access::sqlite-session
         this
         (result columns)
         (unless
           columns
           (set! columns
                 (string*->string-list column-names argc)))
         (push! result (string*->string-list argv argc))
         #f)))

(define-generic
  (execute::bool self::sqlite-session)
  (with-access::sqlite-session
    self
    (connection statement result query)
    (let ((may-be-bound (or query statement)))
         (with-access::sqlite-connection
           connection
           (impl)
           (unless
             may-be-bound
             (error "sqlite-session-execute"
                    "no statement to execute"
                    self))
           (set! result '())
           (let* ((may-be-bound::string may-be-bound)
                  (errmsg::string (pragma::string "NULL"))
                  (errcode
                    (pragma::int
                      "sqlite_exec($1,$2,sqlite_execute_callback,$3, &$4)"
                      impl
                      may-be-bound
                      self
                      errmsg)))
                 (unless
                   (=fx errcode 0)
                   (let ((errmsg
                           (if (pragma::bool "$1 == NULL" errmsg)
                               (pragma::string
                                 "(char *)sqlite_error_string($1)"
                                 errcode)
                               errmsg)))
                        (error "sqlite_exec" errmsg may-be-bound)))
                 (set! result (reverse! result))
                 #t)))))

(define-generic
  (fetch!::pair-nil self::sqlite-session)
  (with-access::sqlite-session
    self
    (result)
    (unless
      result
      (error "session:fetch"
             "Invalid cursor state: not executed"
             self))
    (if (pair? result) (pop! result) '())))

(define-generic
  (fetch-all!::pair-nil self::sqlite-session)
  (with-access::sqlite-session
    self
    (result)
    (begin0 result (set! result #f))))

;(define-generic
;  (_dismiss! self::sqlite-session)
;  (cancel! self))

(define-generic
  (_dismiss! self::sqlite-connection)
  (with-access::sqlite-connection
    self
    (impl)
    (pragma "sqlite_close($1)" impl)
    #unspecified))

(define-generic
  (begin-transaction!::bool
    self::sqlite-connection
    .
    timeout)
  #f)

(define-generic
  (cancel! self::sqlite-session)
  (with-access::sqlite-session
    self
    (result)
    (set! result #f)))

(define-generic
  (has-answer?::bool self::sqlite-session)
  (with-access::sqlite-session
    self
    (result)
    result))

(define-generic
  (bind! self::sqlite-session bindings::pair-nil)
  (with-access::sqlite-session
    self
    (result statement query)
    (set! query (rdbms-format statement bindings))))

(define-generic
  (describe self::sqlite-session)
  (with-access::sqlite-session
    self
    (columns)
    (unless columns (execute self))
    (map list columns)))

